"""
Модуль содержит функцию calc_NPSHA для расчёта надкавитационного напора (кавитационного запаса)
-------------------------------------------------
'Инженерные расчёты на Python'
https://zen.yandex.ru/id/5f33dcd5554adc5b33aaee83
https://medsv.github.io/
"""

from region1 import Region1
from math import pi

__author__ = "Sergey Medvedev"
__copyright__ = "Sergey Medvedev, 2020"
__license__ = "GPL"
__version__ = "1.0"
__maintainer__ = "Sergey Medvedev"
__email__ = "medsv@yandex.ru"
__status__ = "Production"

r1 = Region1()
g = 9.80665  # ускорение свободного падения, м/с2


def calc_NPSH(z_s, z_p, t, p=101325, Q=0, D=0, dH=0):
    """
    Расчёт имеющегося надкавитационного напора NPSHA (кавитационного запаса)
    z_s: высотная отметка поверхности воды, м
    z_p: высотная отметка оси входного патрубка насоса, м
    t: температура воды, С
    p: давление над поверхностью воды, Па. По умолчанию 101325 Па.
    Q: объёмный расход, м3/ч. По умолчанию 0 
    (при расчёте NPSH скорость потока не учитывается)
    D: внутренний диаметр входного патрубка насоса, м. По умолчанию 0
    (при расчёте NPSH скорость потока не учитывается)
    dH: потери давления во входном трубопроводе насоса, м. По умолчанию 0
    (при расчёте NPSH потери давления не учитывается)
    return: значение NPSH, м
    """

    dens = 1 / r1.props_tp(t, p)['v']  # плотность воды при температуре t и давлении p, кг/м3
    H_s = r1.sc.p_t(t) / dens / g  # давление кипения воды при температуре t, м
    H_amb = p / dens / g  # давление над поверхностью воды, м
    if Q > 0 and D > 0:
        if D > 2:  # диаметр указан в мм, поэтому будет выдана ошибка
            raise ValueError("Значение диаметра должно быть задано в метрах")
        w = Q / (pi * D * D / 4) / 3600
    else:
        w = 0.
    NPSH = H_amb + (z_s - z_p) - dH + w * w / 2 / g - H_s
    return NPSH



